#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/sched.h>

#include <linux/kthread.h>
#include <linux/net.h>
#include <net/sock.h>
#include <linux/socket.h>
#include <linux/netlink.h>
#include <linux/skbuff.h>

DEFINE_MUTEX(mut);

#define MAX_PAYLOAD	1024

#define KNETLINK_UNIT 17
#define KNETLINK_GROUP	1234

static struct sock * knetlink_sk = NULL;
u32 pid = 0;

static void netlink_ib_open(void)
{
    struct sk_buff *skb = NULL;
    struct nlmsghdr *nlh = NULL;
    int err;

    skb = alloc_skb(NLMSG_SPACE(MAX_PAYLOAD),GFP_KERNEL);
    NLMSG_PUT(skb, pid, 0, 0, NLMSG_SPACE(MAX_PAYLOAD));
    nlh = (struct nlmsghdr *)skb->data;

    nlh->nlmsg_len =  NLMSG_SPACE(MAX_PAYLOAD);
    nlh->nlmsg_pid = 0; // From kernel
    nlh->nlmsg_flags = NLMSG_DONE;

    strcpy(NLMSG_DATA(nlh), "MESSAGE FROM KERNEL :)");
    NETLINK_CB(skb).dst_group = KNETLINK_GROUP;
    NETLINK_CREDS(skb)->pid = 0; // broadcast, hence no pid

    err = netlink_broadcast(knetlink_sk, skb, 0, KNETLINK_GROUP, GFP_KERNEL);
    printk("\n netlink_broadcast returned %d", err);
nlmsg_failure:
    return;
}

int knetlink_process( struct sk_buff * skb, struct nlmsghdr *nlh )
{
    u8 * payload = NULL;
    int   payload_size;
    int   length;
    int   seq;

    pid = NETLINK_CREDS(skb)->pid;

    length = nlh->nlmsg_len;
    seq = nlh->nlmsg_seq;
    printk("\nknetlink_process: nlmsg len %d type %d pid %d seq %d\n",
	    length, nlh->nlmsg_type, pid, seq );
    /* process the paylad */
    payload_size = nlh->nlmsg_len - NLMSG_LENGTH(0);
    if ( payload_size > 0 ) {
	payload = NLMSG_DATA( nlh );
	printk("\nknetlink_process: Payload is %s ", payload);
    }
    netlink_ib_open();

    return 0;
}

void knetlink_input( struct sk_buff * skb)
{
    mutex_lock(&mut);
    printk("\nFunction %s() called", __FUNCTION__);

    netlink_rcv_skb(skb, &knetlink_process);

    mutex_unlock(&mut);
}

int knetlink_init( void )
{
    if ( knetlink_sk != NULL ) {
	printk("knetlink_init: sock already present\n");
	return 1;
    }

    knetlink_sk = netlink_kernel_create(&init_net, KNETLINK_UNIT, KNETLINK_GROUP, knetlink_input, NULL, THIS_MODULE);
    if ( knetlink_sk == NULL ) {
	printk("knetlink_init: sock fail\n");
	return 1;
    }

    printk("knetlink_init: sock %p\n", (void*)knetlink_sk );
    return 0;
}

void knetlink_exit( void )
{
    if ( knetlink_sk != NULL ) {
	printk("knetlink_exit: release sock %p\n", (void*)knetlink_sk);
	sock_release( knetlink_sk->sk_socket );
    } else {
	printk("knetlink_exit: warning sock is NULL\n");
    }

}

module_init( knetlink_init );
module_exit( knetlink_exit );

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Prashant Bhole");
MODULE_DESCRIPTION("Netlink Demo");
